# postpone evaluation of annotations
from __future__ import annotations

import abc

from typing import Optional, Any, Type, Union, List, AnyStr

from .base import Super, AdHoc

from .utils import get_next_free_po

from pathlib import Path

from base64 import b64encode, b64decode

from re import sub

from queue import Queue

from segno import make_qr, QRCode


def _write(data: AnyStr, *args, **kwargs) -> None:
    po = kwargs.pop("po", None)

    if po:
        mode = "a" if po.exists() else "x"
        kwargs.update({"file": po.open(mode=mode)})

    print(data, *args, **kwargs)

    if po:
        kwargs["file"].close()


def _write_qr(
    qr: QRCode,
    **kwargs,
) -> None:
    po = kwargs.pop("po", None)

    if not po:
        raise ValueError("QRCodeWriter._write_qr: a path object is required!")

    if po.suffix == "":
        po = po.with_suffix(".png")

    po = get_next_free_po(po)

    qr.save(
        po.absolute(),
        dark=kwargs.get("dark") or "black",
        light=kwargs.get("light") or "white",
        scale=kwargs.get("scale") or 20,
        **kwargs,
    )


class Writer(abc.ABC):
    """Derived classes of this base class should implement
    logic for essentially writing data to some buffer,
    preparing that data for writing, and also decoding it
    when reading it out of some buffer.

    Encoding and decoding thus, is not to be taken in the
    str and bytes sense, of encoding to bytes and decoding
    from bytes to str, and instead in the sense of encoding
    arbitrary data in some way, and decoding it back to some
    format.

    We specify that each function has to at least take
    one data argument. Everything further is laissez faire
    including narrowing and expanding arguments in derived
    classes to enforce behaviour.
    """

    @abc.abstractmethod
    def write(self, data: Any, *args, **kwargs) -> None:
        pass

    @abc.abstractmethod
    def encode(self, data: Any, *args, **kwargs) -> Any:
        pass

    @abc.abstractmethod
    def decode(self, data: Any, *args, **kwargs) -> Any:
        pass


class PlainWriter(Writer):
    def write(self, data: AnyStr, *args, **kwargs) -> None:
        _write(data, *args, **kwargs)

    def encode(self, data: Any) -> Any:
        return data

    def decode(self, data: Any) -> Any:
        return data


class HexWriter(Writer):
    def write(self, data: AnyStr, *args, **kwargs) -> None:
        _write(data, *args, **kwargs)

    def encode(self, data: bytes) -> str:
        return data.hex()

    def decode(self, data: str) -> bytes:
        return bytes.fromhex(data)


class Base64Writer(Writer):
    character_encoding: str = "ascii"

    def write(self, data: AnyStr, *args, **kwargs) -> None:
        _write(data, *args, **kwargs)

    def encode(self, data: bytes) -> str:
        return b64encode(data).decode(self.character_encoding)

    def decode(self, data: str) -> bytes:
        return b64decode(data.encode(self.character_encoding))


class FormatWriter(Writer):
    format_template_base: str
    decode_template_pattern: str
    format_template: str = None
    character_encoding: str = "ascii"

    def __format_guard(self) -> None:
        if self.format_template is None:
            raise ValueError("Format template has not been set, can't en- or decode!")

    def write(self, data: AnyStr, *args, **kwargs) -> None:
        _write(data, *args, **kwargs)

    @abc.abstractmethod
    def encode(self, data: Any, *args, **kwargs) -> Any:
        self.__format_guard()

    @abc.abstractmethod
    def decode(self, data: Any, *args, **kwargs) -> Any:
        self.__format_guard()

    @abc.abstractmethod
    def set_format(self, inserts: List[str]) -> str:
        pass


class PEMWriter(FormatWriter):
    format_template_base: str = "-----BEGIN {}-----\n\n{}\n\n-----END {}-----\n"
    decode_template_pattern: str = "-+(BEGIN|END) [ A-Z]+-+"

    @Super.PreCall
    def encode(self, data: bytes) -> str:
        encoded64: str = b64encode(data).decode(self.character_encoding)
        return self.format_template.format(
            "\n".join(
                [
                    encoded64[index * 64 : (index + 1) * 64]
                    for index in range(0, int(len(encoded64) / 64) + 1)
                ]
            )
        )

    @Super.PreCall
    def decode(self, data: str) -> bytes:
        return b64decode(
            "".join(sub(self.decode_template_pattern, "", data).strip()).encode(
                self.character_encoding
            )
        )

    def set_format(self, inserts: List[str]) -> None:
        self.format_template = self.format_template_base.format(
            inserts[0], "{}", inserts[1]
        )


class QRCodeWriter(Base64Writer):
    def write(self, data: AnyStr, *args, **kwargs) -> None:
        _write_qr(make_qr(data, error=kwargs.get("error") or "H", *args), **kwargs)

    def encode(self, data: bytes, **kwargs) -> str:
        return super().encode(data)

    def decode(self, data: str, **kwargs) -> bytes:
        return super().decode(data)


class Printer(Queue):
    _writer_type: Type[Writer]

    writer: Writer

    def __class_getitem__(cls: Type[Self], writer_type: Type[Writer]):
        class ConcretePrinterType(cls):
            pass

        ConcretePrinterType._writer_type = writer_type
        return ConcretePrinterType

    def __init__(self, maxsize=0):
        super().__init__(maxsize)
        self.writer = self._writer_type()

    def print(self, *args, **kwargs) -> None:
        while not self.empty():
            self.writer.write(self.get(), *args, **kwargs)

    @AdHoc.ListMorph(0)
    def put(
        self,
        item: Any,
        encode: bool = True,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> None:
        item = self.writer.encode(item) if encode else item
        super().put(item, block=block, timeout=timeout)



import abc

from typing import Any, List, Callable, Optional

from pydantic import BaseModel


class Super:
    @staticmethod
    def PreCall(func):
        def PrecallsSuper(self, *args, **kwargs):
            getattr(super(type(self), self), func.__name__)(*args, **kwargs)
            return func(self, *args, **kwargs)

        return PrecallsSuper


class AdHoc:
    @staticmethod
    def ListMorph(index) -> Callable[..., Any]:
        def Decorator(func) -> Callable[..., Any]:
            def Morphed(self, *args, **kwargs) -> Any:
                results: List[Any] = []
                if isinstance(args[index], list):
                    for val in args[index]:
                        results.append(
                            func(
                                self, *args[0:index], val, *args[index + 1 :], **kwargs
                            )
                        )
                    return results
                else:
                    return func(self, *args, **kwargs)

            return Morphed

        return Decorator



import click

from pathlib import Path

from cryptoolz.utils import process_pem_file, process_naked_b64

from cryptoolz.app.run import ArgumentDummy, FileArgumentDummy, ToolRunner

from .cryptoolz import cryptoolz, help

_keys_help_text = "This tool creates keypairs in encrypted format \
with different formats to choose from, currently. In future, this tool \
will also allow different algorithms and additional key types. \
\n\n Note that this is an ALPHA version, most errors are not handled, \
meaning that if you did something wrong, then python will weirdly error \
out instead of informing you on what exactly happened. We made sure that \
you can't error on some basic things."


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    help=_keys_help_text,
)
@click.pass_context
def keys(ctx):
    pass


_keys_create_help_text = "This command creates a keypair for a chosen network and encrypts the secret \
with a chosen algorithm combination (designated by the keyword shown in options), formats it into a \
chosen format, and also writes (depending on format or choice etc.) it into a file or stdout. \
Note that the options may also be appended at the end, but it is advisable to prepend them before commands.\
\n\n NUMBER_PASSPHRASES is the number of passphrases used for key creation. We demand either 1 passphrase \
for all keys, or multiple passphrases, for the SAME NUMBER_KEYS_CREATED."


@click.command(help=_keys_create_help_text)
@click.argument(
    "number_passphrases",
    type=int,
    required=False,
    default=1,
    nargs=1,
)
@click.argument(
    "number_keys_created",
    nargs=1,
    required=False,
    type=int,
    default=1,
)
@click.option(
    "-n",
    "--network",
    nargs=1,
    required=False,
    type=click.Choice(["ethereum"]),
    default="ethereum",
    help="The network to create keys for.",
)
@click.option(
    "-a",
    "--algorithm",
    nargs=1,
    required=False,
    type=click.Choice(["aesgcm"]),
    default="aesgcm",
    help="The keyword for the set of algorithms which should be used. For example, aesgcm is currently pbdkf2 + aesgcm.",
)
@click.option(
    "-f",
    "--format",
    nargs=1,
    type=click.Choice(["pem", "qr"]),
    required=False,
    default="pem",
    help="The default output format for encrypted data.",
)
@click.option(
    "-o",
    "--outfile",
    nargs=1,
    type=click.Path(
        exists=False, file_okay=True, dir_okay=False, writable=True, readable=True
    ),
    required=False,
    default=None,
    help="The file to print the created data to.",
)
@click.option(
    "--header",
    nargs=1,
    type=str,
    required=False,
    default=None,
    help="The custom header to use for PEM blocks.",
)
@click.option(
    "--digest/--no-digest",
    type=bool,
    required=False,
    default=False,
    help="Before encryption, compute the digest of the private key. Currently only a dummy option. NOTE: Do NOT send this with encrypted text.",
)
def create(
    number_passphrases,
    number_keys_created,
    network,
    algorithm,
    format,
    outfile,
    header,
    digest,
):
    if digest:
        click.echo("keys :: digest is a dummy option, development ongoing!")
        raise click.Abort

    if format == "qr":
        if not outfile:
            click.echo("keys :: QR code requires a filepath!")
            raise click.Abort
        if header:
            click.echo("keys :: QR code does not take a header arg!")
            raise click.Abort

    if number_passphrases != 1:
        if number_passphrases != number_keys_created:
            click.echo(
                "keys :: If more than one passphrase, the number of keys must be equal to the number of passphrases!"
            )
            raise click.Abort

    runner = ToolRunner()

    pass_dummy = ArgumentDummy(f"bpass:utf-8:{number_passphrases}:True:a passphrase")

    runner.run(
        [pass_dummy],
        t="keys",
        mode="create",
        outpath=Path(outfile) if outfile else None,
        algorithm=algorithm,
        network=network,
        digest=digest,
        format=format,
        header=header or "KEYS",
        keynum=number_keys_created,
    )


_keys_reveal_help_text = 'This command is used to decrypt the encrypted format you have received, \
as output of the `create` command into some file or stdout. This format must be pasted into preferably \
ONE file and then given as input to the command, with the right options, according to how you encrypted \
your data.\n\nNOTE that for QR codes, the QR code must be scanned by the user and only the "plaintext" \
cyphertext should be pasted into a file, see the docs for more.\n\nNUMBER_PASSPHRASES is the number of \
passphrases. If this is set to 1, then the passphrase provided will be used on all encrypted files provided. \
Otherwise, the passphrases provided sequentially will also be used in the same sequence to decrypt the data, \
exactly like during encryption.\n\nFILEPATHS are the paths to the files which contain the key data.'


@click.command(help=_keys_reveal_help_text)
@click.argument(
    "number_passphrases",
    nargs=1,
    type=int,
    default=1,
)
@click.argument(
    "filepaths",
    nargs=-1,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, writable=True, readable=True
    ),
    required=True,
)
@click.option(
    "-n",
    "--network",
    nargs=1,
    type=click.Choice(["ethereum"]),
    required=False,
    default="ethereum",
    help="The network the keys belong to.",
)
@click.option(
    "-a",
    "--algorithm",
    nargs=1,
    type=click.Choice(["aesgcm"]),
    required=False,
    default="aesgcm",
    help="The keyword for the set of algorithms which the data is encrypted with.",
)
@click.option(
    "-f",
    "--format",
    nargs=1,
    type=click.Choice(["pem", "qr"]),
    required=True,
    help="Format of the formatted input cyphertext. In future will be automatic.",
)
@click.option(
    "-o",
    "--outfile",
    nargs=1,
    type=click.Path(
        exists=False, file_okay=True, dir_okay=False, writable=True, readable=True
    ),
    required=False,
    default=None,
    help="The file to print the decrypted data to.",
)
@click.option(
    "--digest/--no-digest",
    type=bool,
    required=False,
    default=False,
    help="Verify private key digest during decryption. Currently a dummy option.",
)
def reveal(number_passphrases, filepaths, network, algorithm, format, outfile, digest):
    if number_passphrases < len(filepaths):
        click.echo(
            "keys :: Can't have more files than passphrases, remove empty files!"
        )
        raise click.Abort

    runner = ToolRunner()

    pass_dummy = ArgumentDummy(
        f"bpass:utf-8:{number_passphrases}:False:the decryption passphrase"
    )
    file_dummy = FileArgumentDummy("file:-1:str:ascii")

    if format == "pem":
        file_dummy._lambda = process_pem_file
    elif format == "qr":
        file_dummy._lambda = process_naked_b64

    runner.run(
        [pass_dummy, file_dummy],
        t="keys",
        mode="reveal",
        f=[Path(p) for p in filepaths],
        outpath=Path(outfile) if outfile else None,
        algorithm=algorithm,
        network=network,
        digest=digest,
        format=format,
    )


keys.add_command(create)
keys.add_command(reveal)
keys.add_command(help)
cryptoolz.add_command(keys)



from typing import List, Any

from cryptoolz.app.tools.base import Args, Options, PreDirector

from cryptoolz.crypto import SecretBytes

from secrets import token_hex

# Cryptograhic circuits and data models which hold our input and output data which are transformed later on
from cryptoolz.crypto.circuits import (
    EncryptPBDKF2_AESGCM,
    DecryptPBDKF2_AESGCM,
    OutsEncryptAESGCM,
    OutsDecryptAESGCM,
)

# For private to public key calc (gives same results as piper's eth_account.Account.from_key)
from eth_hash.auto import keccak
from coincurve import PublicKey
from eth_utils.address import to_checksum_address as checksum

# An instance factory
from cryptoolz.factory import Factory

# Printer class
from cryptoolz.printer import Printer

# The base keypair
from ..base import Keypair

# For picking values out from below
from cryptoolz.utils import match_value, match_values, get_next_free_po

from cryptoolz.crypto.exceptions import WrongDecryptionInputs


FORMATS = {
    "CREATE": {
        "PEM": [
            [
                "\nYour {} Public Key:\n\n{}",
                "",
            ],
            [
                "\nYour {} Public Key:\n\n{}\n",
                "Private key digest:\n\n{}\n",
            ],
        ],
        "QR": [
            "\nYour {} Public Key:\n\n{}\n",
        ],
    },
    "REVEAL": "\nThe decrypted private key (note it down!):\n\n{}\n",
}


def create_exception_handler(e: Exception) -> None:
    raise e


def reveal_exception_handler(e: Exception) -> None:
    if isinstance(e, WrongDecryptionInputs):
        print("Decryption unsuccessful! Either the passphrase or settings are wrong!")
        exit(1)
    else:
        raise e


def get_eth_encryptor_aesgcm():
    return Factory[OutsEncryptAESGCM](EncryptPBDKF2_AESGCM(), create_exception_handler)


def get_eth_decryptor_aesgcm():
    return Factory[OutsDecryptAESGCM](DecryptPBDKF2_AESGCM(), reveal_exception_handler)


class EthKeypair(Keypair):  # TODO: pk digest, trash eth_hash
    def calc_public_key(self, private_key: SecretBytes) -> str:
        if private_key is None:
            return None

        hex_addr = PublicKey.from_valid_secret(private_key.get_secret_value()).format(
            compressed=False
        )[1:]

        return str(checksum(keccak(hex_addr)[-20:].hex()))

    def gen_private_key(self) -> SecretBytes:
        return SecretBytes(bytes.fromhex(token_hex(32)))


class EthKeys:
    director: PreDirector

    def __init__(self, director: PreDirector):
        self.director = director

    def run(self) -> Any:
        command = self.director.options["mode"]
        if command == "create":
            return self.create()
        elif command == "reveal":
            return self.reveal()

    def create(self) -> Any:
        encrypt_outs = []
        options = self.director.options
        results = None

        # Encrypt
        if options["algorithm"] == "aesgcm":
            encrypt_outs = self._encrypt_aesgcm(options)

        # Print
        if options["format"] == "pem":
            results = self._print_create_pem(options, encrypt_outs)
        elif options["format"] == "qr":
            results = self._print_create_qr(options, encrypt_outs)
        return results

    def reveal(self) -> Any:
        results = None
        decrypt_outs = []
        options = self.director.options
        printer = self.director.get_instance(f"{options['format']}_printer")

        # Decode
        decoder = printer.writer
        self.director.args[1] = [
            decoder.decode(cyphertext) for cyphertext in self.director.args[1]
        ]

        # Decrypt
        if options["algorithm"] == "aesgcm":
            decrypt_outs = self._decrypt_aesgcm(options)

        # Print
        return self._print_reveal_plain(options, decrypt_outs)

    def _encrypt_aesgcm(self, options: Options) -> Any:
        keypairs = [EthKeypair() for _ in range(options["keynum"])]
        encryptor = get_eth_encryptor_aesgcm()
        passphrases = self.director.args[0]
        outputs = []

        for i in range(options["keynum"]):
            encryptor.put(
                {
                    "pbdkf2_passphrase": passphrases[i],
                    "aesgcm_plaintext": keypairs[i].private_key,
                }
            )

            outputs.append(
                [
                    keypairs[i].public_key,
                    keypairs[i].private_key_digest,
                    encryptor.create().aesgcm_cyphertext,
                ]
            )
        return outputs

    def _decrypt_aesgcm(self, options: Options) -> Any:
        passphrases = self.director.args[0]
        cyphertexts = self.director.args[1]
        decryptor = get_eth_decryptor_aesgcm()
        outputs = []

        for i in range(options["keynum"]):
            decryptor.put(
                {
                    "pbdkf2_passphrase": passphrases[i],
                    "aesgcm_cyphertext": cyphertexts[i],
                }
            )

            outputs.append(decryptor.create().aesgcm_plaintext)
        return outputs

    def _print_create_pem(self, options: Options, encrypt_outs: List[Any]) -> Any:
        formats = FORMATS["CREATE"]["PEM"]
        formats = formats[options["digest"]]
        network = options["network"][0] + options["network"][1:].lower()
        printer = self.director.get_instance("pem_printer", options["header"])

        for i in range(options["keynum"]):
            printer.put(
                [
                    formats[0].format(network, encrypt_outs[i][0]),
                    formats[1].format(encrypt_outs[i][1]),
                    "The encrypted private key block:\n",
                ],
                encode=False,
            )
            printer.put(encrypt_outs[i][2])

        if not options.get("get"):
            po = None
            if options.get("outpath"):
                po = get_next_free_po(options["outpath"])

            printer.print(po=po)
        else:
            results = []
            while not printer.empty():
                results.append(printer.get())
            return results
        return None

    def _print_create_qr(self, options: Options, encrypt_outs: List[Any]) -> Any:
        formats = FORMATS["CREATE"]["QR"][0]
        network = options["network"][0] + options["network"][1:].lower()
        plain_printer = self.director.get_instance("plain_printer")
        qr_printer = self.director.get_instance("qr_printer")

        for i in range(options["keynum"]):
            plain_printer.put(formats.format(options["network"], encrypt_outs[i][0]))
            qr_printer.put(encrypt_outs[i][0].encode("ascii"), encode=False)
            if options["digest"]:
                qr_printer.put(encrypt_outs[i][1])
                plain_printer.put(
                    encrypt_outs[i][1],
                )
            qr_printer.put(encrypt_outs[i][2])

        if not options.get("get"):
            qr_printer.print(po=options["outpath"])
            plain_printer.print()
        else:
            result = []
            for i in range(options["keynum"]):
                results.append(plain_printer.get())
                results.append(qr_printer.get())
            return results
        return None

    def _print_reveal_plain(self, options: Options, decrypt_outs: List[Any]) -> Any:
        formats = FORMATS["REVEAL"]
        printer = self.director.get_instance("plain_printer")

        printer.put(
            [
                formats.format(plaintext_pk.get_secret_value().hex())
                for plaintext_pk in decrypt_outs
            ]
        )

        if not options.get("get"):
            po = None
            if options["outpath"]:
                po = get_next_free_po(options["outpath"])
            printer.print(po=po)
        else:
            result = []
            for i in range(options["keynum"]):
                results.append(printer.get())
            return results
        return None


def get_eth_keys(director: PreDirector):
    return EthKeys(director)


# And now this can be called from top level

PreDirector.register_builder("ethereum", get_eth_keys)



import abc

# Some types we need for type hints
from typing import List, Optional, Any

from secrets import token_hex

# We use this to manage complexity at least to a certain level
from cryptoolz.app.tools.base import PreDirector

# For the base Keypair
from pydantic import SecretBytes, BaseModel

# Object which formats and sends output to either file or stdout
from cryptoolz.printer import Printer, PlainWriter, HexWriter, PEMWriter, QRCodeWriter


def handle_mode_opt(mode: str):
    if not mode:
        raise ValueError("keys.run: No command to run picked!")
    return mode


PreDirector.register_opt_handlers(
    "default",
    {
        "_get": lambda get: get or False,
        "mode": handle_mode_opt,
        "outpath": lambda outpath: outpath,
        "algorithm": lambda algorithm: algorithm,
        "network": lambda network: network,
        "digest": lambda digest: digest,
        "format": lambda format: format,
        "header": lambda header: header and header.upper(),
        "keynum": lambda keynum: keynum,
    },
)


def handle_passphrases_arg(options, args):
    passphrases = args[0]

    n_passphrases = len(passphrases)

    single_passphrase = n_passphrases == 1

    if len(args) == 2:
        options["keynum"] = len(args[1])

    if options.get("keynum"):
        if not single_passphrase and options["keynum"] != n_passphrases:
            print(
                "keys :: There must be either one passphrase for each encrypted key or the number of passphrases must equal the number of keys!"
            )
            exit(1)
    else:
        options["keynum"] = n_passphrases

    if single_passphrase:
        single_pass_value = passphrases[0]

        for _ in range(1, options["keynum"]):
            passphrases.append(single_pass_value)
    return passphrases


PreDirector.register_arg_handlers("default", [handle_passphrases_arg, None])


def build_plain_printer():
    return Printer[PlainWriter]()


def build_hex_printer():
    return Printer[HexWriter]()


def build_pem_printer(*args):
    printer = Printer[PEMWriter]()
    if 0 < len(args):
        printer.writer.set_format([args[0]] * 2)
    else:
        printer.writer.set_format([""] * 2)
    return printer


def build_qr_printer():
    return Printer[QRCodeWriter]()


PreDirector.register_builder("plain_printer", build_plain_printer)
PreDirector.register_builder("hex_printer", build_hex_printer)
PreDirector.register_builder("pem_printer", build_pem_printer)
PreDirector.register_builder("qr_printer", build_qr_printer)


class Keypair(BaseModel, abc.ABC):
    private_key: SecretBytes
    private_key_digest: Optional[bytes] = None
    public_key: str

    def __init__(self, hash_fn: Any = None, **kwargs):
        private_key = self.gen_private_key()

        private_key_digest = None

        if hash_fn:
            private_key_digest = self.calc_private_key_digest(private_key, hash_fn)

        public_key = self.calc_public_key(private_key)

        super().__init__(
            private_key=private_key,
            private_key_digest=private_key_digest,
            public_key=public_key,
        )

    @abc.abstractmethod
    def calc_public_key(self, private_key: SecretBytes) -> None:
        pass

    @abc.abstractmethod
    def gen_private_key(self) -> SecretBytes:
        pass

    def calc_private_key_digest(
        self, private_key: SecretBytes, hash_fn: Any
    ) -> Optional[bytes]:
        return None



from typing import List, Union

import click

from cryptoolz.crypto.constants import Primes, PRIMES_MERSENNE, PRIMES_LEAST_POW_TWO

from .cryptoolz import cryptoolz, help

_primes_help_text = "Get a prime number according to arguments \
passed to options. Currently only returns primes close to powers of two \
and mersenne primes, and it also doesn't do probabilistic checks, it \
instead just reads the data from included files."


@click.command(help=_primes_help_text)
@click.option(
    "--least",
    "prime",
    flag_value="least",
    required=False,
    default=True,
    help="Primes of the form 2^n - k. This is the default.      n ∈ ([8,400],512), 0 <= k < (n != 512 ? 10 : 1)",
)
@click.option(
    "--mersenne",
    "prime",
    flag_value="mersenne",
    required=False,
    help="Primes of the form 2^p - 1, where p is prime.          2 <= p < 20000",
)
@click.option(
    "--check/--no-check",
    type=bool,
    required=False,
    default=False,
    help="Just print all the possible values for the components of the selected prime type \
(such as n, k, p). Ignores --bits and --rank.",
)
@click.option(
    "-b",
    "--bits",
    nargs=1,
    type=int,
    required=False,
    default=None,
    help="Supply the number of bits, meaning which power to raise 2 to.",
)
@click.option(
    "-r",
    "--rank",
    nargs=1,
    type=int,
    required=False,
    default=-1,
    help="Supply the rank meaning either the index of the mersenne prime if 8 is at 0, or \
the index of the k value for the least 2^n - k prime.",
)
def primes(prime, check, bits, rank):
    missing = lambda prime: f"primes :: no arguments supplied for --{prime}"
    both = (
        lambda prime, first, second: f"primes :: --{prime} requires both --{first} and --{second}"
    )
    takes_one = (
        lambda prime, first, second: f"primes :: --{prime} takes only one of --{first}, --{second}"
    )

    try:
        if prime == "mersenne":
            if check:
                click.echo(PRIMES_MERSENNE)
            elif bits and 0 <= rank:
                click.echo(takes_one(prime, "bits", "rank"))
            elif bits:
                click.echo(Primes.get_mersenne(bits))
            elif 0 <= rank:
                click.echo(Primes.get_mersenne_by_rank(rank))
            else:
                click.echo(missing(prime))
                exit(1)
        elif prime == "least":
            if check:
                click.echo(PRIMES_LEAST_POW_TWO)
            elif bits and 0 <= rank:
                click.echo(Primes.get_least_near_pow_two(bits, rank))
            elif bits or 0 <= rank:
                click.echo(both(prime, "bits", "rank"))
            else:
                click.echo(missing(prime))
                exit(1)
    except Exception as e:
        print(e)
        exit(1)


cryptoolz.add_command(primes)



from typing import Callable, Any


class ArgumentDummy:
    kind: str

    def __init__(self, kind: str):
        if kind is None:
            raise ValueError("ArgumentDummy.__init__: Kind can not be None!")

        self.kind = kind


class FileArgumentDummy(ArgumentDummy):
    _lambda: Callable[..., Any]

    def __init__(self, kind: str):
        self._lambda = lambda input: input
        super().__init__(kind)

    def get_lambda(self) -> Callable[..., Any]:
        return self._lambda



"""Console script for cryptoolz."""
from typing import List, Tuple, Any

import sys
import click

from pydantic import SecretBytes, SecretStr


def handle_click_opts(
    *arguments: Tuple[Any], **opts_dict: dict[str, List[Any]]
) -> dict[str, Any]:
    for key, val in opts_dict.items():
        for nested in val:
            if nested in arguments:
                opts_dict[key] = arguments[arguments.index(nested) + 1]

    return opts_dict


class SecretStrClickParamType(click.ParamType):
    name = "secret string"

    def convert(self, value, param, ctx):
        try:
            return SecretStr(str(value))
        except:
            self.fail(
                f"cryptoolz :: click: {value} cannot be converted to a string!",
                param,
                ctx,
            )


class SecretBytesClickParamType(click.ParamType):
    name = "secret string"
    encoding: str

    def __init__(self, encoding: str | None):
        if encoding is None:
            raise ValueError(
                "cryptoolz :: SecretBytesClickParamType.__init__: encoding can't be none!"
            )
        self.encoding = encoding

    def convert(self, value, param, ctx):
        try:
            return SecretBytes(str(value).encode(self.encoding))
        except:
            self.fail(
                f"cryptoolz :: click: {value} cannot be converted to a string!",
                param,
                ctx,
            )


ClickSecretStr = SecretStrClickParamType()
ClickSecretBytesASCII = SecretBytesClickParamType("ascii")
ClickSecretBytesUTF8 = SecretBytesClickParamType("utf-8")


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option("-m", "--module", type=str, nargs=1, required=False)
@click.pass_context
def cryptoolz(ctx, module):
    pass


@click.command(help="Show this message and exit.")
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())


cryptoolz.add_command(help)



from typing import Any
from ..base import Args, Options
from .base import PreDirector
from .dropins import *


class Keys:
    @staticmethod
    def run(*args: Args, **options: Options) -> Any:
        director = PreDirector(options.get("network"), args, options)
        runner = director.get_instance(options.get("network"), director)
        return runner.run()



from pathlib import Path
from cryptoolz.utils import ROOT_DIR
from importlib import import_module

dropins_po = Path(ROOT_DIR + "/app/tools/keys/dropins")

__all__ = []

for dropin_po in dropins_po.iterdir():
    if dropin_po.name != "__init__.py" and dropin_po.name != "__pycache__":
        import_module(
            "cryptoolz.app.tools.keys.dropins." + dropin_po.with_suffix("").stem
        )
        __all__.append(dropin_po.with_suffix("").stem)



import os
import json

from typing import List, Dict, Union, Optional, Any

from pathlib import Path

from re import finditer


def process_pem_file(
    pem_encoded: List[str], context_chunks: List[Union[str, List[str]]]
) -> List[str]:
    results = []

    for pem_text in pem_encoded:
        matches = [match for match in finditer(r"-+(BEGIN|END) [ A-Z]+-+", pem_text)]

        if len(matches) % 2 != 0:
            raise RuntimeError(
                f"{context_chunks[0]} :: FileDummy._lambda: File {context_chunks[1][pem_encoded.index(pem_text)]} has an odd number of begin and end lines!"
            )

        for i in range(0, len(matches), 2):
            results.append(pem_text[matches[i].start() : matches[i + 1].end()])

    return results


def process_naked_b64(
    b64_lines_block: List[str], context_chunks: List[Union[str, List[str]]]
) -> List[str]:
    results = []

    for block in b64_lines_block:
        lines = block.split("=")
        for i in range(len(lines) - 1):
            results.append(lines[i].strip() + "=")

    return results


def match_value(cases: Dict[str, Any], value: Any) -> Any:
    if value in cases:
        cases = cases[value]
    else:
        raise ValueError("utils.match_value :: matched is None!")
    return cases


def match_values(cases: Dict[str, Any], *values: List[Any]) -> Any:
    for value in values:
        if value in cases:
            cases = cases[value]
        else:
            raise ValueError("utils.match_values :: matched is None!")
    return cases


def get_next_free_po(po: Path) -> Path:
    counter = 1
    stem = po.stem
    while po.exists():
        po = po.with_stem(f"{stem}{counter}")
        counter += 1
    return po


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))  # cryptoolz root


def get_resource(filename: str) -> Any:
    with (Path(ROOT_DIR) / "resources" / filename).open(mode="r") as fo:
        return json.load(fo)
    raise RuntimeError("utils.get_resource: should have never reached this point!")



import abc

from typing import List, Union, Optional, Type, TypeVar, Generic

# Circuits are Base Models because circuit data is important
from pydantic import BaseModel, validator
from pydantic import Field, SecretBytes, SecretField

from pydantic.generics import GenericModel
from pydantic.errors import MissingError

from secrets import token_bytes

from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC as pbdkf2
from cryptography.hazmat.primitives.ciphers.aead import AESGCM as aesgcm
from cryptography.hazmat.primitives.hashes import (
    Hash,
    SHA3_512 as sha3_512,
    BLAKE2b as blake2b,
)

from cryptography.exceptions import InvalidTag

from .exceptions import WrongDecryptionInputs

# Bases


class CircuitInsBase(BaseModel):
    class Config:
        # Validate field defaults
        validate_all = True
        # Validate fields on assignment
        validate_assignment = True


class CircuitOutsBase(BaseModel):
    class Config:
        # Fields are immutable
        allow_mutation = False


CircuitInsType = TypeVar("InType", bound=CircuitInsBase)
CircuitOutsType = TypeVar("OutType", bound=CircuitOutsBase)


class CircuitLike(GenericModel, Generic[CircuitInsType, CircuitOutsType], abc.ABC):
    ins: CircuitInsType
    outs: Optional[CircuitOutsType] = None

    @abc.abstractmethod
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, *args, **kwargs) -> CircuitOutsType:
        if len(kwargs) != 0:
            self.ins = type(self.ins)(**kwargs)
        self.outs = type(self).compute_outs(self.ins)
        return self.outs

    @abc.abstractclassmethod
    def validate_circuit_ins(cls, ins: CircuitInsType) -> CircuitInsType:
        pass

    @abc.abstractclassmethod
    def compute_outs(cls, ins: CircuitInsType) -> CircuitOutsType:
        pass

    class Config:
        # Validate field defaults
        validate_all = True
        # Validate fields on assignment
        validate_assignment = True


SecretType = TypeVar("SecretType", bound=SecretField)

InternalInput = Union
ExternalInput = Union[SecretType, None]
PublicOutput = Union
SecretOutput = Union[SecretType, None]

# BLAKE2b


class InsBLAKE2B(BaseModel):
    blake2b_data: ExternalInput[List[SecretBytes]] = None
    # Set to True to get intermediate digests
    blake2b_intermediate: InternalInput[bool] = False
    # Currently only 64 supported
    blake2b_digest_size: InternalInput[int] = 64


class OutsBLAKE2B(BaseModel):
    blake2b_digests: PublicOutput[List[bytes]] = []


class HashBLAKE2B(CircuitLike[InsBLAKE2B, OutsBLAKE2B]):
    def __init__(self, **kwargs):
        super().__init__(ins=InsBLAKE2B(**kwargs))

    @validator("ins")
    def validate_circuit_ins(cls, ins: InsBLAKE2B) -> InsBLAKE2B:
        if ins.blake2b_digest_size != 64:
            raise ValueError(
                "HashBLAKE2B.validate_circuit_ins: digest size may currently only be 64!"
            )
        return ins

    @classmethod
    def compute_outs(cls, ins: InsBLAKE2B) -> OutsBLAKE2B:
        if ins.blake2b_data is None or len(ins.blake2b_data) == 0:
            raise ValueError("HashBLAKE2B.compute_outs: There is not data to hash!")

        digests = []
        hash_fn = Hash(blake2b(64))

        if ins.blake2b_intermediate:
            for bytes_data in ins.blake2b_data:
                digests.append(hash_fn.update(bytes_data).copy().finalize())
        else:
            for bytes_data in ins.blake2b_data:
                hash_fn.update(bytes_data)

        digests.append(hash_fn.finalize())

        return OutsBLAKE2B(blake2b_digests=digests)


# PBDKF2HMAC

pbdkf2_accepted_digests = {
    "sha3_512": sha3_512,
}

pbdkf2_max_key_size_bytes = {
    "sha3_512": (2**32 - 1) * pbdkf2_accepted_digests["sha3_512"].digest_size
}


class InsPBDKF2(BaseModel):
    pbdkf2_passphrase: ExternalInput[SecretBytes] = None
    pbdkf2_hash_fn_name: InternalInput[str] = "sha3_512"
    pbdkf2_key_size_bytes: InternalInput[int] = 32
    pbdkf2_salt: InternalInput[bytes] = Field(default_factory=lambda: token_bytes(32))
    pbdkf2_iterations: InternalInput[int] = 40000
    pbdkf2_salt_size_bytes: InternalInput[int] = 32


class OutsPBDKF2(BaseModel):
    pbdkf2_derived_key: SecretOutput[SecretBytes] = None
    pbdkf2_salt: PublicOutput[bytes] = b""


class InsDerivePBDKF2(CircuitInsBase, InsPBDKF2):
    pass


class OutsDerivePBDKF2(CircuitOutsBase, OutsPBDKF2):
    pass


class DerivePBDKF2(CircuitLike[InsDerivePBDKF2, OutsDerivePBDKF2]):
    def __init__(self, **kwargs):
        super().__init__(ins=InsDerivePBDKF2(**kwargs))

    @validator("ins")
    def validate_circuit_ins(cls, ins: InsDerivePBDKF2) -> InsDerivePBDKF2:
        salt_len = len(ins.pbdkf2_salt)

        # General
        if ins.pbdkf2_hash_fn_name not in pbdkf2_accepted_digests:
            raise ValueError(
                f"DerivePBDKF2.compute_outs: {ins.pbdkf2_hash_fn_name} is not part of this module because it is not recommended, or does not exist."
            )
        if ins.pbdkf2_salt_size_bytes != salt_len:
            raise ValueError(
                "DerivePBDKF2.validate_circuit_ins: Salt size does not correspond to stated size!"
            )

        # Minimums
        if salt_len < 16:
            raise ValueError(
                "DerivePBDKF2.validate_circuit_ins: Salt should be minimum 16 bytes!"
            )
        if ins.pbdkf2_key_size_bytes < 16:
            raise ValueError(
                "DerivePBDKF2.validate_circuit_ins: Key size should be minimum 16 bytes!"
            )
        if ins.pbdkf2_iterations < 10000:
            raise ValueError(
                "DerivePBDKF2.validate_circuit_ins: Iterations should be minimum 10000!"
            )

        # Maximums
        if (
            pbdkf2_max_key_size_bytes[ins.pbdkf2_hash_fn_name]
            < ins.pbdkf2_key_size_bytes
        ):
            raise ValueError(
                "DerivePBDKF2.validate_circuit_ins: derived key size is above maximum allowed according to digest!"
            )

        return ins

    @classmethod
    def compute_outs(cls, ins: InsDerivePBDKF2) -> OutsDerivePBDKF2:
        if ins.pbdkf2_passphrase is None:
            raise ValueError(
                "DerivePBDKF2.compute_outs: A derivation passphrase is necessary!"
            )

        digest = pbdkf2_accepted_digests[ins.pbdkf2_hash_fn_name]

        kdf = pbdkf2(
            digest, ins.pbdkf2_key_size_bytes, ins.pbdkf2_salt, ins.pbdkf2_iterations
        )

        return OutsDerivePBDKF2(
            pbdkf2_derived_key=SecretBytes(
                kdf.derive(ins.pbdkf2_passphrase.get_secret_value())
            ),
            pbdkf2_salt=ins.pbdkf2_salt,
        )


# AESGCM

aesgcm_nonce_max_len = (2**64 - 1) / 8


class InsAESGCM(BaseModel):
    aesgcm_key: ExternalInput[SecretBytes] = None
    aesgcm_plaintext: ExternalInput[SecretBytes] = None
    # IV should be at least 12 bytes, and is usually set to 12 bytes as such
    aesgcm_nonce: InternalInput[bytes] = Field(default_factory=lambda: token_bytes(12))


class OutsAESGCM(BaseModel):
    # For decryption
    aesgcm_key: SecretOutput[SecretBytes] = None
    # The cyphertext has the nonce concatenated
    aesgcm_cyphertext: PublicOutput[bytes] = b""
    # If the nonce is not 12 bytes, set this
    aesgcm_nonce_size_bytes: PublicOutput[int] = 12


class InsEncryptAESGCM(CircuitInsBase, InsAESGCM):
    pass


class OutsEncryptAESGCM(CircuitOutsBase, OutsAESGCM):
    pass


class EncryptAESGCM(CircuitLike[InsEncryptAESGCM, OutsEncryptAESGCM]):
    def __init__(self, **kwargs):
        super().__init__(ins=InsEncryptAESGCM(**kwargs))

    @validator("ins")
    def validate_circuit_ins(cls, ins: InsEncryptAESGCM) -> InsEncryptAESGCM:
        key_len = ins.aesgcm_key and len(ins.aesgcm_key)
        nonce_len = len(ins.aesgcm_nonce)

        # General
        if key_len and key_len % 16 != 0:
            if key_len != 24:
                raise ValueError(
                    f"EncryptAESGCM.validate_circuit_ins: Key must be either 16, 24 or 32 bytes, not {key_len}!"
                )

        # Minimums
        if nonce_len < 12:
            raise ValueError(
                "EncryptAESGCM.validate_circuit_ins: Nonce should be minimum 12 bytes!"
            )

        # Maximums
        if aesgcm_nonce_max_len < nonce_len:
            raise ValueError(
                "EncryptAESGCM.validate_circuit_ins: Nonce exceeds size limit of 2**64 - 1 bits!"
            )

        return ins

    @classmethod
    def compute_outs(cls, ins: InsEncryptAESGCM) -> OutsEncryptAESGCM:
        if ins.aesgcm_key is None:
            raise ValueError(
                "EncryptAESGCM.compute_outs: Cannot encrypt without an AESGCM key!"
            )
        if ins.aesgcm_plaintext is None:
            raise ValueError("EncryptAESGCM.compute_outs: No plaintext to encrypt!")

        return OutsEncryptAESGCM(
            aesgcm_key=None,
            aesgcm_cyphertext=aesgcm(ins.aesgcm_key.get_secret_value()).encrypt(
                ins.aesgcm_nonce, ins.aesgcm_plaintext.get_secret_value(), None
            )
            + ins.aesgcm_nonce,
        )


class InsDecryptAESGCM(CircuitInsBase, OutsAESGCM):
    pass


class OutsDecryptAESGCM(CircuitOutsBase, InsAESGCM):
    pass


class DecryptAESGCM(CircuitLike[InsDecryptAESGCM, OutsDecryptAESGCM]):
    def __init__(self, **kwargs):
        super().__init__(ins=InsDecryptAESGCM(**kwargs))

    @validator("ins")
    def validate_circuit_ins(cls, ins: InsDecryptAESGCM) -> InsDecryptAESGCM:
        return ins

    @classmethod
    def compute_outs(cls, ins: InsDecryptAESGCM) -> OutsDecryptAESGCM:
        if ins.aesgcm_key is None:
            raise ValueError(
                "DecryptAESGCM.compute_outs: Cannot decrypt without an AESGCM key!"
            )
        if ins.aesgcm_cyphertext == b"":
            raise ValueError("DecryptAESGCM.compute_outs: No cyphertext to decrypt!")

        aesgcm_nonce: bytes = ins.aesgcm_cyphertext[-ins.aesgcm_nonce_size_bytes :]

        try:
            return OutsDecryptAESGCM(
                aesgcm_plaintext=aesgcm(ins.aesgcm_key.get_secret_value()).decrypt(
                    aesgcm_nonce,
                    ins.aesgcm_cyphertext[: -ins.aesgcm_nonce_size_bytes],
                    None,
                ),
                aesgcm_nonce=aesgcm_nonce,
            )
        except InvalidTag as ite:
            raise WrongDecryptionInputs(
                "DecryptAESGCM.compute_outs: Decryption inputs are wrong!"
            ) from ite

        raise RuntimeError(
            "DecryptAESGCM.compute_outs: program should have never reached this point!"
        )


# combined


class InsEncryptPBDKF2_AESGCM(CircuitInsBase, InsAESGCM, InsPBDKF2):
    pass


class EncryptPBDKF2_AESGCM(CircuitLike[InsEncryptPBDKF2_AESGCM, OutsEncryptAESGCM]):
    def __init__(self, **kwargs):
        super().__init__(ins=InsEncryptPBDKF2_AESGCM(**kwargs))

    @validator("ins")
    def validate_circuit_ins(
        cls, ins: InsEncryptPBDKF2_AESGCM
    ) -> InsEncryptPBDKF2_AESGCM:
        EncryptAESGCM.validate_circuit_ins(DerivePBDKF2.validate_circuit_ins(ins))
        return ins

    @classmethod
    def compute_outs(cls, ins: InsEncryptPBDKF2_AESGCM) -> OutsEncryptAESGCM:
        if ins.pbdkf2_passphrase is None:
            raise ValueError(
                "EncryptPBDKF2_AESGCM.compute_outs: A derivation passphrase is necessary!"
            )
        if ins.aesgcm_plaintext is None:
            raise ValueError(
                "EncryptPBDKF2_AESGCM.compute_outs: No plaintext to encrypt!"
            )

        pbdkf2_result: OutsPBDKF2 = DerivePBDKF2.compute_outs(ins)

        ins.aesgcm_key = pbdkf2_result.pbdkf2_derived_key

        aesgcm_result: OutsEncryptAESGCM = EncryptAESGCM.compute_outs(ins)

        return OutsEncryptAESGCM(
            aesgcm_key=None,
            aesgcm_cyphertext=aesgcm_result.aesgcm_cyphertext + ins.pbdkf2_salt,
        )


class InsDecryptPBDKF2_AESGCM(CircuitInsBase, OutsAESGCM, InsPBDKF2):
    pass


class DecryptPBDKF2_AESGCM(CircuitLike[InsDecryptPBDKF2_AESGCM, OutsDecryptAESGCM]):
    def __init__(self, **kwargs):
        super().__init__(ins=InsDecryptPBDKF2_AESGCM(**kwargs))

    @validator("ins")
    def validate_circuit_ins(
        cls, ins: InsDecryptPBDKF2_AESGCM
    ) -> InsDecryptPBDKF2_AESGCM:
        DecryptAESGCM.validate_circuit_ins(DerivePBDKF2.validate_circuit_ins(ins))
        return ins

    @classmethod
    def compute_outs(cls, ins: InsDecryptPBDKF2_AESGCM) -> OutsDecryptAESGCM:
        if ins.pbdkf2_passphrase is None:
            raise ValueError(
                "DecryptPBDKF2_AESGCM.compute_outs: A passphrase for key derivation is necessary!"
            )

        ins.pbdkf2_salt = ins.aesgcm_cyphertext[-ins.pbdkf2_salt_size_bytes :]

        pbdkf2_result: OutsPBDKF2 = DerivePBDKF2.compute_outs(ins)

        return DecryptAESGCM.compute_outs(
            InsDecryptAESGCM(
                aesgcm_key=pbdkf2_result.pbdkf2_derived_key,
                aesgcm_cyphertext=ins.aesgcm_cyphertext[: -ins.pbdkf2_salt_size_bytes],
            )
        )



from __future__ import annotations

import abc

from typing import List, Dict, Union, Callable, TypeVar, Any


# Our args will be in list format
Args = List[Any]

# Our options (from kwargs) are in a dict
Options = Dict[str, Any]

# TypeVar for generics
SingleType = TypeVar("SingleType")

# Our registry separates based on command name
Registry = Dict[str, SingleType]

# Handlers are functions that handle opt values
OptHandlers = Dict[str, Callable[List[Any], Any]]

# The registry stores for each command opts
OptHandlerRegistry = Registry[OptHandlers]

# Arg handlers are functions that handle args
ArgHandlers = List[Callable[List[Any], Any]]

# Arg handler list registry with int key
ArgHandlerRegistry = Registry[ArgHandlers]

# Builders build any class instances
BuilderDict = Dict[str, Callable[..., Any]]

# Registry for them
BuilderRegistry = Registry[BuilderDict]


class PreDirector:
    # static
    _opt_handlers_registry: OptHandlerRegistry = {}
    _arg_handlers_registry: ArgHandlerRegistry = {}
    _builder_registry: BuilderRegistry[Options] = {}

    # dynamic
    args: Args
    options: Options

    def __init__(self, name: str, args: Args, options: Options):
        self.args = [None] * len(args)

        for key, value in options.items():
            handler = self._opt_handlers_registry.get(name)
            handler = handler and handler.get(key)
            if handler:
                options.update({key: handler(value)})
            else:
                options.update(
                    {key: self._opt_handlers_registry["default"][key](value)}
                )

        self.options = options

        for i in range(len(args)):
            handler = self._arg_handlers_registry.get(name)
            handler = handler and handler.get(i)
            if not handler:
                handler = self._arg_handlers_registry["default"][i]
            if handler:
                self.args[i] = handler(self.options, args)
            else:
                self.args[i] = args[i]
        self.name = name

    def get_instance(self, instance_name: str, *args: Args, **options: Options) -> Any:
        return self._builder_registry[instance_name](*args, **options)

    @classmethod
    def register_builder(cls, key: str, builder: Callable[..., Any]) -> None:
        cls._builder_registry.update({key: builder})

    @classmethod
    def register_opt_handlers(cls, key: str, handlers: OptHandlers) -> None:
        cls._opt_handlers_registry.update({key: handlers})

    @classmethod
    def register_arg_handlers(cls, key: str, handlers: ArgHandlers) -> None:
        cls._arg_handlers_registry.update({key: handlers})



from typing import Any, Optional, Union, List, Callable

# For processing filepaths
from pathlib import Path

# Function to get user input from console
from getpass import getpass

# Using secret values for input as much as possible
from pydantic import SecretBytes, SecretStr

# These dummies represent arguments which should be
# received from the user or files dinamically
from .dummies import ArgumentDummy, FileArgumentDummy

# Import all tool modules
from cryptoolz.app.tools import Tools


class ToolRunner:
    """
    This class is an opt handler and tool executor, so a bridge between the cli
    and the tools.

    :param get_lambdas: holds lambdas which are used for processing files
        or getting user input, this is such that one may modify them on the go.
    """

    get_lambdas: dict[str, Callable[..., Any]]

    def __init__(self) -> None:
        # Default lambdas
        self.get_lambdas = {
            "int": lambda format, encoding: input(format),
            "str": lambda format, encoding: input(format),
            "pass": lambda format, encoding: SecretStr(getpass(format)),
            "bytes": lambda format, encoding: input(format).encode(encoding),
            "bpass": lambda format, encoding: SecretBytes(
                getpass(format).encode(encoding)
            ),
        }

    def run(self, run_args: List[Any], **run_options) -> Any:
        """This prepares data for and runs a tool module by passing all
        the arguments and options it requires into it's constructor (run_options)
        and run function (run_args).

        :param run_args: The arguments for running the tool wanted by the `t` run
            option. `run_args` can also contain argument dummies, which it recognizes
            and consumes, replacing them with the actual arguments wanted. This is
            the way we prompt for stuff like passwords, because we prefer exerting
            control over such behaviour instead of letting click handle it.
        :param run_options: kwargs-like run options which are passed into the
            tool's constructor, there are two important ones:
            t - This is the tool being called.
            f - These are filepaths which complement the file argument dummy.
            We pass them in as such because of convenience, might move
            this into FileDummy later on.
        """
        # Gets
        tool: Optional[str] = run_options.get("t")
        filepaths: Optional[List[Path]] = run_options.get("f") or None

        # Checks
        if not tool:
            raise ValueError(
                f"{tool} :: ToolRunner.run: A tool to execute must be specified!"
            )

        # Pop ToolRunner options
        run_options.pop("t", None)
        run_options.pop("f", None)

        # Replaces Dummies with user inputted data
        for i in range(len(run_args)):
            argument = run_args[i]

            if isinstance(argument, ArgumentDummy):
                kind_chunks = argument.kind.split(":")

                if kind_chunks[0] == "file":
                    run_args[i] = argument._lambda(
                        self.get_file_data(kind_chunks, [tool, filepaths]),
                        [tool, filepaths],
                    )
                elif kind_chunks[0] in self.get_lambdas:
                    run_args[i] = self.get_user_supplied(kind_chunks, [tool])
                else:
                    raise ValueError(
                        f"{tool} :: ToolRunner.run: There is no lambda for this kind!"
                    )

        # run tool
        return Tools[tool].run(*run_args, **run_options)

    def get_user_supplied(
        self, kind_chunks: List[str], context_chunks: List[str]
    ) -> Any:
        """
        A function which returns user inputs based on a kind and context format.
        The format for user inputted is as follows:
            kind_chunks:
                kind:encoding:size:confirm:printed
            context_chunks:
                tool
        """
        result: Any = []

        _lambda = self.get_lambdas[kind_chunks[0]]

        size = int(kind_chunks[2])

        print_suffix = lambda index: f" at index [{i}]: " if size != 1 else ": "

        for i in range(0, size):
            result.append(
                _lambda(
                    f"{context_chunks[0]} :: please enter {kind_chunks[4]}{print_suffix(i)}",
                    kind_chunks[1],
                )
            )

            if kind_chunks[3] == "True":
                if result[-1] != _lambda(
                    f"{context_chunks[0]} :: please repeat to confirm: ",
                    kind_chunks[1],
                ):
                    print(f"{context_chunks[0]} :: The inputs must equal, exiting!")
                    exit(1)

        return result

    def get_file_data(
        self, kind_chunks: List[str], context_chunks: List[Union[str, List[Path]]]
    ) -> List[Union[str, bytes]]:
        """
        A function which returns file data based on a kind, lambda and context format.
        The format for processing files is as follows:
            kind_chunks:
                kind:maxread:type:encoding
            context_chunks:
                tool:filepaths

        Argument dummies have to implement a file handling lambda if kind is file.
        """
        results = []

        for po in context_chunks[1]:
            po = po if str(po)[0] == "/" else po.absolute()

            if po.exists():
                with open(str(po), mode="r") as fo:
                    results.append(fo.read(int(kind_chunks[1])))
            else:
                print(
                    f"{context_chunks[0]} :: ToolRunner.get_file_data: Couldn't find {str(po)}! Exiting!"
                )
                exit(1)

        for i in range(0, len(results)):
            if kind_chunks[2] == "bytes":
                results[i] = results[i].encode(kind_chunks[3])

        return results

    def add_lambda(self, key: str, lambda_to_add: Callable[..., Any]):
        self.get_lambdas.update({key: lambda_to_add})

    def remove_lambda(self, key: str):
        self.get_lambdas.pop(key, None)



from pydantic import SecretBytes
import cryptoolz.crypto.exceptions as exceptions
import cryptoolz.crypto.constants as constants
import cryptoolz.crypto.circuits as circuits
import cryptoolz.crypto.data as data

__all__ = [
    "SecretBytes",
    "exceptions",
    "constants",
    "circuits",
    "data",
]



# postpone evaluation of annotations
from __future__ import annotations

from typing import Any, Generic, TypeVar, Optional, Callable, Generator, List

from .base import AdHoc

from queue import Queue

FactoryItem = TypeVar("FactoryItem")


class Factory(Queue, Generic[FactoryItem]):
    constructor: Callable[..., FactoryItem]

    _handle_exception: Callable[[Exception], None]

    defaults: Optional[dict[str, Any]] = None

    stored: List[FactoryItem] = []

    def __init__(
        self,
        constructor: Callable[..., FactoryItem],
        exception_handler: Callable[[Exception], None] = None,
        defaults: Optional[dict[str, Any]] = None,
        maxsize=0,
    ):
        super().__init__(maxsize)

        self.constructor = constructor

        self._handle_exception = exception_handler

        if defaults:
            self.defaults = defaults

    def stream(self, some: int = 0) -> Generator[FactoryItem, None, None]:
        qsize: int = self.qsize()
        some = some or qsize
        predicate = min if qsize else max

        for _ in range(predicate(qsize, some)):
            yield self.create()

    def create(self, store: bool = False) -> FactoryItem:
        result = None

        data_dict = self.defaults if self.empty() else self.get()

        args = data_dict.get("args") or []

        data_dict.pop("args", None)

        try:
            result = self.constructor(*args, **data_dict)
        except Exception as e:
            if self._handle_exception:
                self._handle_exception(e)
            else:
                raise e

        if store:
            self.stored.append(result)

        return result

    @AdHoc.ListMorph(0)
    def put(
        self,
        item: Any,
        block: bool = True,
        timeout: Optional[float] = None,
    ):
        super().put(item, block=block, timeout=timeout)



from typing import Callable
from cryptoolz.utils import get_resource

# FORMAT: { 'n': ['k'] }
# PRIME: 2**n - k
PRIMES_LEAST_POW_TWO = get_resource("primes_least_pow_two.json")

# FORMAT: [p]
# PRIME: 2**p - 1
PRIMES_MERSENNE = [
    2,
    3,
    5,
    7,
    13,
    17,
    19,
    31,
    61,
    89,
    107,
    127,
    521,
    607,
    1279,
    2203,
    2281,
    3217,
    4253,
    4423,
    9689,
    9941,
    11213,
    19937,
]


class Primes:
    @staticmethod
    def get_least_near_pow_two(bitnum: int, rank: int):
        """Supports 8-400 and 512 bits, currently."""
        power = str(bitnum)
        ks = PRIMES_LEAST_POW_TWO.get(power)
        if ks:
            if rank < len(ks):
                return 2**bitnum - int(ks[rank])
            else:
                raise ValueError(
                    f"Primes.get_least_near_pow_two: in 2^{power} - k_i, i<={rank} not supported."
                )
        else:
            raise ValueError(f"Primes.get_least_near_pow_two: 2^{power} not supported.")

    @staticmethod
    def get_mersenne_by_rank(rank: int):
        if rank < len(PRIMES_MERSENNE):
            return Primes.get_mersenne(PRIMES_MERSENNE[rank])
        else:
            raise ValueError(
                f"Primes.get_mersenne_by_rank: rank <= {rank} not supported."
            )

    @staticmethod
    def get_mersenne(bitnum: int):
        """Primes below 20k bits supported."""
        if bitnum in PRIMES_MERSENNE:
            return 2**bitnum - 1
        else:
            raise ValueError(
                f"Primes.get_mersenne: 2^{bitnum} - 1 is not prime or not supported."
            )



from .circuits import InsEncryptPBDKF2_AESGCM, InsDecryptPBDKF2_AESGCM

get_example_data_constructor = {
    "encryption": {
        "pbdkf2": {
            "aesgcm": lambda: InsEncryptPBDKF2_AESGCM(
                pbdkf2_passphrase="test password".encode("ascii"),
                aesgcm_plaintext=b"1d5d2f4382130e73bc5434c47b32a76a86fb0765a6d578afcf4a57ce3427e868",
                aesgcm_nonce=b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
                pbdkf2_salt=b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
            ).dict()
        }
    },
    "decryption": {
        "pbdkf2": {
            "aesgcm": lambda: InsDecryptPBDKF2_AESGCM(
                pbdkf2_passphrase="test password".encode("ascii"),
                aesgcm_cyphertext=b'-\xd8\x85I=@{\x97\xf7\xad+\x1e\xae\xda\x14@b}\xba\xab\x80s\xed\xab\x1f\x94M\xd1\xf3\x8d&\xa4,\x17\xdc\xe0\x1a\x10Y+\xaf\x00\x99e"\xdf\x98\x8a\xb7\n\x7f\x89\x9a\xac\x17\x8eF \x12\x9a\xd8\xf5g#`i\xae\xcfoB7\x075\xcd\xd5D\xcf\xe39e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
            ).dict()
        }
    },
}
